#' Scorecards function to create multi-system or multi-reference scorecards by 
#' region (types 9 & 11)
#' 
#'@description function to create a scorecard for a multi-system/multi-reference 
#'  combination, showing data by region and forecast month
#'  
#'@param data is an array of spatially aggregated metrics data, containing the
#'   following dimensions; system, reference, metric, region, sdate, ftime.
#'@param system a vector of character strings defining the system following the
#'   archive.yml format from verification suite 
#'@param reference a vector of character strings defining the name of the 
#'  reference following the archive.yml format from verification suite 
#'@param var a character string following the format from 
#'  variable-dictionary.yml from verification suite 
#'@param start.year a numeric indicating the start year of the reference period
#'@param end.year a numeric indicating the end year of the reference period
#'@param start.date a vector of character strings indicating the start months
#'@param forecast.month a vector of numeric indicating the forecast months
#'@param region.names a vector of character strings containing names of the 
#'  regions corresponding to the input data
#'@param metrics a vector of character strings containing the metrics,  
#'  NOTE: must be in same order as in data.
#'@param scorecard.type a numeric indicating the type of scorecard to plot
#'@param fileout a path of the location to save the scorecard plots
#'  
#'@example 
#'scorecard_plot <- SC_plot_simple_by_region(data = aggregated_metrics,
#'                                           system = 'ecmwfs5' ,
#'                                           reference = 'era5',
#'                                           var = 'tas',
#'                                           start.year = 1993,
#'                                           end.year = 2016,
#'                                           start.months = 1:12
#'                                           forecast.months = 1:6,
#'                                           region.names = c('Tropics', 'Extra-tropical NH', 'Extra-tropical SH')
#'                                           metrics = c('mean_bias', 'enscorr', 'rpss','crpss', 'enssprerr'),
#'                                           scorecard.type = 9,
#'                                           fileout = '/esarchive/scratch/nmilders/scorecards_data/saved_results'
#'                                           )

SC_plot_multi_by_region <- function(data,
                                     system,
                                     reference,
                                     var,
                                     start.year,
                                     end.year,
                                     start.months,
                                     forecast.months,
                                     region.names,
                                     metrics,
                                     scorecard.type,
                                     fileout){
  
  # 
  # ## For testing:
  # data <- aggregated_metrics
  # system <- c('ecmwfs5', 'dwds21')
  # reference <-'era5'
  # var <- 'tas'
  # start.year <- 1993
  # end.year <- 2016
  # start.months <- sprintf("%02d", 1:12)
  # forecast.months <- 1:6
  # region.names <- c('Tropics', 'Extra-tropical NH', 'Extra-tropical SH')
  # metrics <- c('mean_bias', 'enscorr', 'rpss','crpss', 'enssprerr')
  # scorecard.type <- 9
  # fileout <- '/esarchive/scratch/nmilders/scorecards_data/saved_results'
  
  ## Load configuration files
  archive <- read_yaml("./conf/archive.yml")$archive
  var_dict <- read_yaml("./conf/variable-dictionary.yml")$vars
  
  ## Round data
  data <- round(data,2)
  
  ## get dimensions of inputs
  n.start.months <- as.numeric(dim(data)["sdate"]) ## TO DO: create sdate_dim
  n.ftime.months <- as.numeric(dim(data)["time"]) ## TO DO: create ftime_dim
  n.metrics <- length(metrics) 
  n.regions <- length(region.names) 
  n.systems <- length(system)
  n.references <- length(reference)
  
  ## Can only show multiple systems with one reference or multiple references 
  ## with one system, need to select which
  

  ## Get long names for showing in table
  var.name <- var_dict[[var]]$long_name
  var.units <- var_dict[[var]]$units
  
  # for(sys in n.systems){
  #   system.name <- archive$System[[which(archive$System %like% system[sys])]]$system_name ## need to do in loop??
  # }
  # for(ref in n.references){
  #   reference.name <- archive$Reference[[which(archive$Reference %like% reference[ref])]]$reference_name ## need to do in loop??
  # }
  
  ## Temporary! TO DO: get names for vector, or do loop?
  system.name <- system
  reference.name <- reference
  
  metric.names.list <- list(mean_bias = paste0("Mean bias (", var.units,")"),
                            enscorr = "Correlation",
                            rpss = "RPSS",
                            crpss = "CRPSS",
                            enssprerr = "Spread-to-error ratio")
  for(met in metrics){
    if(met == metrics[1]){
      metric.names <- metric.names.list[[met]] 
    } else {
      metric.names <- c(metric.names, metric.names.list[[met]])
    }
  }
  
  
  if(n.systems > 1 && n.references == 1){
    model <- 'System'
    n.models <- n.systems
    eval.label <- 'Ref'
    eval.name <- reference.name
  }
  if(n.systems == 1 && n.references > 1){
    model <- 'Reference'
    n.models <- n.references
    model.col.header <- 'References'
    eval.label <- 'Sys'
    eval.name <- system.name
  }
  
  ## Define table size
  n.rows <- n.models * n.ftime.months 
  n.columns <- 2 + (n.metrics * n.start.months)
  
  ## Column names for table
  column.names.figure <- rep("", n.models * n.ftime.months)
  system.name.figure <- gsub("System ", "S", system.name)
  for(ss in 1:n.models) column.names.figure[floor(n.ftime.months/2) + (ss-1) * n.ftime.months ] = system.name.figure[ss]

  ## Define scorecard table titles
  column.names <- c(model, "Forecast month", rep(c(month.abb[as.numeric(start.months)]), n.metrics))
  
  if(scorecard.type == 9){
    column.header <- "Start date"
  }
  if(scorecard.type == 11){
    column.header <- "Target month"
  }

  ## Loop over region if more than one region is requested
    for(reg in 1:n.regions){
      
      options(stringsAsFactors = FALSE) 
      table.title <- data.frame(c1=paste0(var.name, ".  Region: ", region.names[reg]), c2=n.columns)
      header.names <- as.data.frame(data.frame(c1=c("", metric.names), c2=c(2, rep(n.start.months, n.metrics)))) ## TO DO: need to include metric units
      header.names2 <- as.data.frame(data.frame(c1=c("", paste0(rep(column.header, n.metrics))), c2=c(2, rep(n.start.months, n.metrics))))
      ref.label <- data.frame(c1=paste0("(", eval.label, ": ", eval.name, " ", start.year, "-", end.year, ")"), c2 = n.columns)
      
      ## Define data inside the scorecards table
      ## NEED TO FIX: allow different or of metrics than in data

      table_model <- NULL
      
        if(model == 'System'){
          for(sys in 1:n.systems){
            table_metric <- data.frame(table_column_2 = as.character(forecast.months))
            for(met in 1:n.metrics){
              table_metric <- data.frame(table_metric, data[sys,1,met,,,reg])  
            }
            table_model <- rbind(table_model, table_metric)
          }
        } ## close if on system
      
        if(model == 'Reference'){
          for(ref in 1:n.references){
            table_metric <- data.frame(table_column_2 = as.character(forecast.months))
            for(met in 1:n.metrics){
              table_metric <- data.frame(table_metric, data[1,ref,met,,,reg])  
            }
            table_model <- rbind(table_model, table_metric)
          }
        } ## close if on reference

        
      ## All data for plotting in table
      table <- data.frame(table_column_1 = column.names.figure, table_model)
      
      ## Find position of mean bias metric to calculate breaks
      if ('mean_bias' %in% metrics) {
        pos_bias <- which(metrics == 'mean_bias')
        breaks_bias <- SC_breaks(Subset(data, along = 'metric', 
                                        indices = pos_bias))
      }
      
      ## Define breaks for each metric based of metric position:
      metric.breaks <- list('mean_bias' = breaks_bias,
                            'enscorr' = c(-1, -0.8, -0.6, -0.4, -0.2, 0, 0.2, 0.4, 0.6, 0.8, 1),
                            'rpss' = c(-0.5, -0.4, -0.3, -0.2, -0.1, 0, 0.1, 0.2, 0.3, 0.4, 0.5),
                            'crpss' = c(-0.5, -0.4, -0.3, -0.2, -0.1, 0, 0.1, 0.2, 0.3, 0.4, 0.5),
                            'enssprerr' = c(0, 0.6, 0.7, 0.8, 0.9, 1, 1.2, 1.4, 1.6, 1.8, 2))
      
      ## Set table background colors
      palette <- c('#2D004B', '#542789', '#8073AC', '#B2ABD2', '#D8DAEB', '#FEE0B6', '#FDB863', '#E08214', '#B35806', '#7F3B08')
      colorunder <- "#04040E"
      colorsup <- "#730C04"
      
      metric.col <- array(colorunder, c(n.ftime.months * n.models, n.columns-2)) 
      metric.text.col <- array("#2A2A2A", c(n.ftime.months * n.models, n.columns-2)) ## black is the default text color ## gray : "#2A2A2A"
      # metric.text.bold <- array(TRUE, c(n.ftime.months * n.models, n.columns-2)) ## Setting all values to bold
      
      ## temporary table in array format
      table_temp <- array(unlist(table[3:n.columns]), dim=c(n.rows,n.columns-2)) ##as.matrix(table[3:n.columns])
      
      ## Define cell and text colors to show in table
      for(met in 1:n.metrics){
        metric.int <- metric.breaks[[metrics[met]]]
        for(rr in 1:n.rows){
          for(stm in 1:n.start.months){
            for(pp in 1:(length(metric.int)-1)){
              if(is.nan(table_temp[rr,((met - 1) * n.start.months + stm)])){
                metric.col[rr,((met - 1) * n.start.months + stm)] <- "gray"
              } else {
                if(table_temp[rr,((met - 1) * n.start.months + stm)] >= metric.int[pp] && table_temp[rr,((met - 1) * n.start.months + stm)] <= metric.int[pp+1])
                  metric.col[rr,((met - 1) * n.start.months + stm)] <- palette[pp]
                
                if(table_temp[rr,((met - 1) * n.start.months + stm)] < metric.int[1])
                  metric.col[rr,((met - 1) * n.start.months + stm)] <- colorunder
                if(table_temp[rr,((met - 1) * n.start.months + stm)] >= metric.int[length(metric.int)])
                  metric.col[rr,((met - 1) * n.start.months + stm)] <- colorsup
              }
              
              ## color text in white and bold if background is white or dark blue or dark red:
              if(is.nan(table_temp[rr,((met - 1) * n.start.months + stm)]) ||
                 (!is.nan(table_temp[rr,((met - 1) * n.start.months + stm)]) && pp == 1 && table_temp[rr,((met - 1) * n.start.months + stm)] < metric.int[2]) ||
                 (!is.nan(table_temp[rr,((met - 1) * n.start.months + stm)]) && pp == 2 && table_temp[rr,((met - 1) * n.start.months + stm)] < metric.int[3]) ||
                 (!is.nan(table_temp[rr,((met - 1) * n.start.months + stm)]) && pp == (length(metric.int) - 1) && table_temp[rr,((met - 1) * n.start.months + stm)] >= metric.int[length(metric.int) - 1]) ||
                 (!is.nan(table_temp[rr,((met - 1) * n.start.months + stm)]) && pp == (length(metric.int) - 2) && table_temp[rr,((met - 1) * n.start.months + stm)] >= metric.int[length(metric.int) - 2]))
              {
                metric.text.col[rr,((met - 1) * n.start.months + stm)] <- "white"
                #metric.text.bold[rr,((met - 1) * n.start.months + stm)] <- TRUE
              }
            } ## close loop on pp
          } ## close loop on stm
        } ## close loop on rr
      } ## close loop on met
      
      
      metric.col.df <- data.frame(metric.col)
      metric.text.col.df <- data.frame(metric.text.col)
      #metric.text.bold.df <- data.frame(metric.text.bold)
      
      ## Create HTML table
      table.html.part <- list()
      table.html.part[[1]] <- kbl(table, escape = F, col.names = column.names, align = rep("c", n.columns)) %>%
        kable_paper("hover", full_width = F) %>%
        add_header_above(header = header.names2, font = 14)%>%
        add_header_above(header = ref.label, align = "left", font_size = 14) %>%
        add_header_above(header = header.names, font_size = 20) %>%
        add_header_above(header = table.title, font_size = 20, align = "left")
        
      for(i in 1:n.metrics){ 
        for(stm in 1:n.start.months){ 
          my.background <- metric.col.df[(i - 1) * n.start.months + stm]
          my.text.color <- metric.text.col.df[(i - 1) * n.start.months + stm ]
          #my.bold <- metric.text.bold.df[(i - 1) * n.start.months + stm]
          
          table.html.part[[(i - 1) * n.start.months + stm + 1]] <-
            column_spec(table.html.part[[(i - 1) * n.start.months + stm]], 2 + n.start.months * (i - 1) + stm, 
                        background = my.background[1:n.rows,], 
                        color = my.text.color[1:n.rows,], 
                        bold = T) ## strsplit(toString(bold), ', ')[[1]]
        } ## close loop on stm
      } ## close loop on i
      
      n.last.list <- n.metrics * n.start.months + 1
      
      table.html <- column_spec(table.html.part[[n.last.list]], 1, bold = TRUE, width_min = "4cm", width_max = "4cm") %>%
        column_spec(1, bold = TRUE, width_min = "4cm", width_max = "4cm") %>% 
        column_spec(2, bold = TRUE, width_max = "1cm") %>% 
        #column_spec(3:n.columns, width_max = "1.2cm", width_min = "1.2cm") %>% ## Columns too small
        
        column_spec(c(1, 2, 2+n.start.months, 2+n.start.months*2, 2+n.start.months*3, 2+n.start.months*4), border_right = "2px solid black") %>% ## Need to make flexible for number of metrics
        ##column_spec(c(1,2,14,26,38,50), border_right = "2px solid black") %>%
        #### use function cumsum ?? 
        
        column_spec(1, border_left = "2px solid black") %>%
        column_spec(n.columns, border_right = "2px solid black") %>%
        row_spec(seq(from = 0, to = n.ftime.months * n.models, by = n.ftime.months), extra_css = "border-bottom: 2px solid black", hline_after = TRUE) #1px solid")
      
      
      
      ## Create color bar legends
      if('mean_bias' %in% metrics){
        png(filename= paste0(fileout, "/scorecard_legend_mean_bias.png"), width=495, height=40) ## width = 455
        ColorBar(brks = metric.breaks[['mean_bias']], cols = palette, vertical = FALSE, triangle_ends_scale = 1, col_inf = colorunder, col_sup = colorsup) 
        dev.off()
      }
      if('enscorr' %in% metrics){
        png(filename= paste0(fileout, "/scorecard_legend_enscorr.png"), width=495, height=40)
        ColorBar(brks = metric.breaks[['enscorr']], cols = palette, vertical = FALSE, triangle_ends_scale = 0, col_inf = colorunder)
        dev.off()
      }
      if('rpss' %in% metrics){
        png(filename= paste0(fileout, "/scorecard_legend_rpss.png"), width=495, height=40)
        ColorBar(brks = metric.breaks[['rpss']], cols = palette, vertical = FALSE, col_inf = colorunder, col_sup = colorsup) 
        dev.off()
      }
      if('crpss' %in% metrics){
        png(filename= paste0(fileout, "/scorecard_legend_crpss.png"), width=495, height=40)
        ColorBar(brks = metric.breaks[['crpss']], cols = palette, vertical = FALSE, col_inf = colorunder, col_sup = colorsup) 
        dev.off()
      }
      if('enssprerr' %in% metrics){
        png(filename= paste0(fileout, "/scorecard_legend_enssprerr.png"), width=495, height=40)
        ColorBar(brks = metric.breaks[['enssprerr']], cols = palette, vertical = FALSE, triangle_ends_scale = 1,  col_sup = colorsup)
        dev.off()
      }
      
      currentworkingdir <- getwd()
      setwd(fileout)
      
      ## Paste all legends together
      for(m in metrics){
        if(m == metrics[1]){
          ## Add white space to the left of the first color bar legend
          system(paste0("convert ./scorecard_legend_", m, ".png -background white -splice 225x0 ./legend_scorecard.png"))
        } else {
          system(paste0("montage ./legend_scorecard.png ./scorecard_legend_",m,".png -tile 2x1 -geometry -4+10 ./legend_scorecard.png"))
        }
      }
      unlink('scorecard_legend*')
      
      ## Save html table as png
      my.file <- paste0(fileout, '/scorecard.png')
      try(save_kable(table.html, file = my.file), silent = TRUE)
      
      
      ## Remove temporal directories (not automatically removed due to error with save_kable)
      #unlink(paste0(fileout, './saved_results/lib*'), recursive = TRUE) # clean temporal directories (not automatically removed because of error with save_kable)
      
      ## Add white space to the top of the scorecards, to be cut when adding legend:
      setwd(fileout)
      system(paste0("convert ./scorecard.png -background white -splice 0x20 ./scorecard.png"))
      
      ## Add the legends below the scorecard table
      system(paste0("montage ./scorecard.png ./legend_scorecard.png -tile 1x2 -geometry +0-10 ./scorecard.png"))
      
      ## Save scorecard png in desired directory
      dirout <- save_scorecard_multi(model, eval.name, var, start.year, end.year, scorecard.type, sub(" ", "-", region.names[reg]))
      system(paste0("mv ", './scorecard.png', paste0(" ", dirout))) 
      print(paste0("Scorecard type ", scorecard.type, " figure saved - Multi-", tolower(model), " (", eval.label, ": ", eval.name, ", Region: ", region.names[reg],")"))
      
      unlink('legend_scorecard.png')
      setwd(currentworkingdir)
  
    } ## close loop on reg
  
}## close function
